__author__ = 'nickmab'

import types
import threadpool
from . import _utils
from _utils import validate_type as _validate_type

"""Contains convenience classes for working with basic ThreadPools and similar basic tasks."""

class FuncWrap:
    """Class that wraps a func call, including args, kwargs, result or exceptions.

    Useful in conjunction with a basic ThreadPool (for example) if you do not wish to bother
    separately implementing a separate callback for handing results and exceptions.

    Attributes:
        target (func): The function to be called. This is the only thing guaranteed to 
            have a True truth value no matter what (i.e. nothing else is required).
        name (str): You can name this FuncWrap instance for future reference.
        args (tuple): The arguments to be provided to the target func when called.
        kwargs (dict): Keyword arguments to be provided to your target func when called.
        result (object): Whatever your function returned (presuming run has been called),
            which could also be None.
        exception (Exception): If an exception was raised during your function's execution,
            it will be here (or None if not).

    Args:
        target (func): The function to be called.
        name (str, optional): You can name this FuncWrap instance for future reference.
        args (tuple, optional): The arguments to be provided to the target func when called.
        kwargs (dict, optional): Keyword arguments to be provided to your target func when called.
    
    Example:
        
        Imagine we have a simple function that we want to run in a separate thread::

            def f(a, b, c=None, d=1):
                x = a + b
                if x:
                    x += c
                return x * d

        We can run this function and capture the results for future reference as follows::

            fw = FuncWrap(target=f, name='example', args=(2, 3), kwargs={ 'c': 4, 'd': 5 })
            fw.run()

        We can subsequently find out what happened::

            # in this case there is no exception, but if there was it would be here...
            if fw.exception:
                # in real life you might not just throw it again, but this is just an example...
                raise fw.exception
            else:
                # the return value will be (2 + 3 + 4) * 5 == 45
                print 'the result is %s' %fw.result

    """
    def __init__(self, target=None, name='', args=(), kwargs={}):
        """init with target function, optional name, args and kwargs to identify and call that function.

        The target function is the only required argument here (i.e. you do not have to name it and of course
        your function might not need any arguments at all).

        Args:
            target (func): The function to be called.
            name (str, optional): You can name this FuncWrap instance for future reference.
            args (tuple, optional): The arguments to be provided to the target func when called.
            kwargs (dict, optional): Keyword arguments to be provided to your target func when called.

        Raises:
            ValueError: If no target function is provided.
            TypeError: If any of the arguments provided are not the correct type.

        """
        # validation first
        if not target:
            raise ValueError('Must provide a valid target function at a minimum.')
        _validate_type('target', target, types.FunctionType)
        _validate_type('name', name, types.StringType)
        _validate_type('args', args, types.TupleType)
        _validate_type('kwargs', kwargs, types.DictionaryType)

        self.target = target 
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.exception = None

    def run(self):
        """Call target func with args and kwargs in a try/except and store the result (or exception)."""
        try:
            self.result = self.target(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e
        

class FuncWrapPool:
    """Runs a collection of FuncWrap instances using workers in a threadpool.ThreadPool.

    Attributes:
        func_wraps (list): This contains the list of FuncWrap instances whose targets are 
            called by worker functions in a ThreadPool. Inspect this list of FuncWraps after
            the results have all returned to get the results and/or exceptions.

    Example:
        Imagine we want to run two long-running functions at ond safely capture results::

            def f(a, b):
                return a + b

            def g(s='hello'):
                return s.count('h')

        We can package these as FuncWraps as follows (in real life you might have a
        long list of different arguments to call these functions for):: 

            # Any iterable will do; we will use a list and just construct these manually inline.
            fws = [FuncWrap(target=f, args=(2, 3)), FuncWrap(target=g, kwargs={ 's': 123 })]

        We can now put these in a FuncWrapPool and run them asynchronously. Note that the second 
        of these FuncWraps will obviously result in an AttributeError:

            fwp = FuncWrapPool(2, fws)
            # a call to run() will block until all results are in
            fwp.run()

        So the FuncWrap instances in our "fws" list (above) will now be populated with results/exceptions::

            res1 = fws[0].result # this will be 2 + 3 == 5 and the exception will be None
            raise fws[1].exception # the exception will be here, the result will be None. 

    """
    def __init__(self, num_workers=5, func_wrap_instances_iterable=[]):
        """Init with a number of worker threads and an iterable of FuncWrap instances to run later.

        Args:
            num_workers (int, optional): The (default 5) number of workers to put in our threadpool.ThreadPool.
            func_wrap_instances_iterable (object): Must be an iterable that yields references (using 
                standard for _ in iter_obj: syntax) to FuncWrap instances for this FuncWrapPool instance
                to run in workers in a threadpool.ThreadPool.

        Raises:
            TypeError: Throws if the func_wrap_instances_iterable argument is not iterable! Also throws
                if num_workers is not an int.
            ValueError: Although it is a kwarg, the default empty list for the func_wrap_instances_iterable
                argument is not acceptable and this will throw a ValueError if it is not provided.

        """
        # validation and logic inline
        if not func_wrap_instances_iterable:
            raise ValueError('Must provide a non-empty iterable with FuncWrap instances!')
        _validate_type('num_workers', num_workers, types.IntType)
        self.num_workers = num_workers
        self.func_wraps = []
        for fw in func_wrap_instances_iterable:
            try:
                _validate_type('func_wrap', fw, FuncWrap)
            except TypeError as e:
                raise TypeError('FuncWrapPool expects an iterable of FuncWrap instances. ' +
                    'Instead got %s as part of %s.' %(fw, func_wrap_instances_iterable))
            self.func_wraps.append(fw)
        
    def run(self):
        """Run all FuncWraps in a threadpool.ThreadPool and wait (block) til all have returned."""
        pool = threadpool.ThreadPool(self.num_workers)
        [pool.putRequest(threadpool.WorkRequest(fw.run)) for fw in self.func_wraps]
        pool.wait()



__author__ = 'nickmab'

import types
from . import _utils
from _utils import validate_type as _validate_type
from _utils import is_coroutine as _is_coroutine
import multiprocessing as mp

"""Contains boilerplate code for working with common subprocess-related tasks."""

def generate_and_consume(producer_gen, consumer_func, print_warns=True):
    """Run (and exhaust) a producer generator and consumer function in two separate processes. 

    Given a generator function that produces objects for a handler function to consume,
    set up a pipe between two subprocesses, then send the objects one at a time to
    the consumer function.

    Args:
        producer_gen (generator): Must yield any non-None object to be passed
            through a multiprocessing.Pipe to a handler function.
        consumer_func (function): Must take two arguments. The first is the object
            produced by the producer in the other process and the second is a
            multiprocessing.Queue to which it can optionally put output or results.
        print_warns (bool, optional): True by default. If an exception is caught
            in either the producer or consumer process, a line will be printed to
            stdout specifying this if set to True. Set False to suppress output.

    Returns:
        multiprocessing.Queue: The results produced by the consumer function.

    Raises:
        ValueError: If the producer generator yields a None.
        TypeError: If any of the arguments are invalid types.
        Exception: Any exceptions other than expected StopIteration raised by the two
            child processes will be caught and raised again in the parent process. 

    Example:
        An example of a generator::

            def gen():
                for i in range(10):
                    yield i

        An example of a handler::

            def handle(obj, out_q):
                # do something with the obj...
                out_q.put(2*obj)

        You would use this as follows::

            res = generate_and_consume(gen(), handle)
            tot = 0
            while not res.empty():
                tot += res.get()

        The total would, of course, be 90.

    """
    _validate_type('producer_gen', producer_gen, types.GeneratorType)
    _validate_type('consumer_func', consumer_func, types.FunctionType)
    _validate_type('print_warns', print_warns, types.BooleanType)
    def p_func(conn):
        try:
            while True:
                obj = producer_gen.next()
                if obj is None:
                    raise ValueError('Producer generator %s produced a "None" value! Invalid.' 
                        %producer_gen.__name__)
                conn.send(obj)
        except Exception as e:
            # 'None' sentinel denotes completion
            conn.send(None)
            if not isinstance(e, StopIteration):
                # this indicates there was actually an error, so let's send it through.
                conn.send(e)
            conn.close()
    
    def c_func(conn, lock, out_q, err_q):
        try:
            while True:
                item = conn.recv()
                if item is None:
                    # 'None' sentinel denotes completion, but check whether there was an exception...
                    if conn.poll(0.01):
                        # if there's something more in the queue, it means we've encountered an exception.
                        with lock:
                            if print_warns:
                                print 'Encountered exception in producer generator %s!' %producer_gen.__name__
                            err_q.put(conn.recv())
                    conn.close()
                    return 

                consumer_func(item, out_q)
        except Exception as e:
            with lock:
                if print_warns:
                    print 'Encountered exception in consumer function %s!' %consumer_func.__name__
                err_q.put(e)
            conn.close()

    conn_1, conn_2 = mp.Pipe()
    lock, err_q, out_q = mp.Lock(), mp.Queue(), mp.Queue()
    p1 = mp.Process(target=p_func, args=(conn_1,))
    p2 = mp.Process(target=c_func, args=(conn_2, lock, out_q, err_q))
    p1.start()
    p2.start()
    # we only need to wait for the consumer process to finish. It keeps going until it gets the 
    # 'None' sentinel from the producer *or* an exception.
    p2.join()
    if p1.is_alive():
        # might still be a live if there's an exception raised in the consumer.
        p1.terminate()
    if not err_q.empty():
        raise err_q.get()
    return out_q

def generate_and_consume_in_coroutine(producer_gen, consumer_coroutine, 
        print_warns=True, trust_coroutine_impl=False):
    """Same as generate_and_consume function, except the consumer is a coroutine.

    If you are not familiar with python coroutines, google them first. A coroutine is
    somewhat like a generator; it can maintain state, keep running indefinitely
    and is written in a style that resembles an ordinary function. They differ in their use
    of the yield keyword. A generator produces values using yield (which are retrieved by
    calling next()) whereas a coroutine receives input using yield (which is sent by calling send()).

    Given a generator function that produces objects for a handler coroutine to consume,
    set up a pipe between two subprocesses, then send the objects one at a time to
    the consumer coroutine.

    Why use a coroutine? So that you can maintain state inside the consumer and
    only produce output values into the queue when it makes sense (instead of having to do it
    from a stateless single function call as in the generate_and_consume function).

    Args:
        producer_gen (generator): Must yield any non-None object to be passed
            through a multiprocessing.Pipe to a handler function.
        consumer_coroutine (function): Must be a function that will return a coroutine 
            when called. The function must take one argument - a multiprocessing.Queue to which it 
            can optionally put output or results. The coroutine must handle all inputs 
            from the yield keyword, must also yield no outputs at all, and must exit 
            when a None object is sent to it. generate_and_consume_in_coroutine will send None as
            a sentinel to indicate that the coroutine should finish putting output in the Queue and return.
        print_warns (bool, optional): True by default. If an exception is caught
            in either the producer or consumer process, a line will be printed to
            stdout specifying this if set to True. Set False to suppress output.
        trust_coroutine_impl (bool, optional): False by default. Set this to True if you know
            that a call to inspect.getsourcelines(consumer_coroutine) will fail to find the source code
            for your coroutine (e.g. if you're running this in an interactive shell) and so will
            not be able to accurately validate it.

    Returns:
        multiprocessing.Queue: The results produced by the consumer coroutine.

    Raises:
        ValueError: If the producer generator yields a None or the consumer coroutine
            does not follow the above description (although whether it correctly handles
            the None sentinel indicating the need to exit is not validated.
        TypeError: If any of the arguments are invalid types.
        Exception: Any exceptions other than expected StopIteration raised by the two
            child processes will be caught and raised again in the parent process. 

    Example:
        An example of a generator::

            def gen():
                for i in range(10):
                    yield i
        
        A valid coroutine consumer::

            def co(out_q):
                tot = 0
                while True:
                    obj = yield
                    if obj is None:
                        out_q.put(tot)
                        return
                    tot += obj

        You would use these as follows::

            # set trust_coroutine_impl to be true if we're running from an interactive shell.
            out_q = generate_and_consume_in_coroutine(gen(), co, trust_coroutine_impl=True)
    
        The out_q would be a queue with one integer, being the sum of the numbers 0 to 9.

    """
    _validate_type('producer_gen', producer_gen, types.GeneratorType)
    _validate_type('consumer_coroutine', consumer_coroutine, types.FunctionType)
    _validate_type('print_warns', print_warns, types.BooleanType)
    if not trust_coroutine_impl and not _is_coroutine(consumer_coroutine):
        raise ValueError('Consumer coroutine %s code is invalid. ' %consumer_coroutine.__name__ +
            'yield keyword must only ever appear alone on a line (to drop sent input) ' + 
            'or to the rhs of an assignment operator.')        

    def p_func(conn):
        try:
            while True:
                obj = producer_gen.next()
                if obj is None:
                    raise ValueError('Producer generator %s produced a "None" value! Invalid.' 
                        %producer_gen.__name__)
                conn.send(obj)
        except Exception as e:
            # 'None' sentinel denotes completion
            conn.send(None)
            if not isinstance(e, StopIteration):
                # this indicates there was actually an error, so let's send it through.
                conn.send(e)
            conn.close()
    
    def c_func(conn, lock, out_q, err_q):
        # try to make our coroutine
        try:
            cons = consumer_coroutine(out_q)
            cons.next()
        except Exception as e:
            with lock:
                if print_warns:
                    print 'Encountered an exception trying to initialise ' + \
                        'consumer coroutine %s!' %consumer_coroutine.__name__
                err_q.put(e)
            conn.close()
            return
        
        try:
            stopiter_expected = False
            while True:
                item = conn.recv()
                if item is None:
                    # 'None' sentinel denotes completion, but check whether there was an exception...
                    if conn.poll(0.01):
                        # if there's something more in the queue, it means we've encountered an exception.
                        with lock:
                            if print_warns:
                                print 'Encountered exception in producer generator %s!' %producer_gen.__name__
                            err_q.put(conn.recv())
                    stopiter_expected = True
                    cons.send(None)
                    return
                cons.send(item)
        except Exception as e:
            if not isinstance(e, StopIteration) or not stopiter_expected:
                with lock:
                    if print_warns:
                        print 'Encountered exception in consumer coroutine %s!' %consumer_coroutine.__name__
                    err_q.put(e)
            conn.close()

    conn_1, conn_2 = mp.Pipe()
    lock, err_q, out_q = mp.Lock(), mp.Queue(), mp.Queue()
    p1 = mp.Process(target=p_func, args=(conn_1,))
    p2 = mp.Process(target=c_func, args=(conn_2, lock, out_q, err_q))
    p1.start()
    p2.start()
    # we only need to wait for the consumer process to finish. It keeps going until it gets the 
    # 'None' sentinel from the producer *or* an exception.
    p2.join()
    if p1.is_alive():
        # might still be a live if there's an exception raised in the consumer.
        p1.terminate()
    if not err_q.empty():
        raise err_q.get()
    return out_q



__author__ = 'nickmab'

"""Module for doing web requests with the help of nickmab.async_util.threads module."""

import types
import urllib2
import json
from . import threads as thr
from _utils import validate_type as _validate_type

class JSONQueryPool(object):
    """Run JSON web requests in a ThreadPool, collect and parse results into a python obj.

    This class actually uses a nickmab.async_util.threads.FuncWrapPool to hand out jobs
    to worker threads and then collect the results. The results will be collected in a dict
    of python objects, being either the parsed JSON result or an exception raised in a
    worker thread.

    Whilst the jobs are run asynchronously, a call to JSONQueryPool.run is blocking.

    Attributes:
        result (dict): The results dict, which will be empty before the workers have returned
            and of the form { 'query_str_label': result_obj } where the label is provided
            per query in the class __init__  

    Examples:
        Let's imagine we have two urls http://foo.bar/bas and http://bas.bar/foo that return
        (as JSON string) a list [1, 2, 3] and a dict { 'a': 'b } respectively. 

        We can fetch these as follows::

            # create a pool with the details - note the query { 'name': 'url' } format 
            qp = JSONQueryPool(num_worker_threads=2, queries={
                'bas_list': 'http://foo.bar.bas', 'foo_dict': 'http://bas.bar/foo' }) 
            
            # run both queries at once and collect the results (this call blocks til done)
            qp.run()

        Now we have our results in the result dict and can fetch them::

            if not qp.has_exception:
                l = qp.result['bas_list']
                d = qp.result['foo_dict']
            else:
                # do some other exception handling...
                for k, v in qp.result.iteritems():
                    if isinstance(v, Exception):
                        print 'Encountered exception in %s query: %s' %(k, v)

    """
    def __init__(self, num_worker_threads=5, queries={}):
        """init by specifying number of worker threads, query names / identifiers and urls.

        Does not run immediately. Some basic validation is performed and we wait for a call to run().

        Args:
            num_worker_threads (int, optional): Number of worker threads in the ThreadPool, default 5.
            queries (dict): { str: str } dict of 'query_name': 'query_url' where the name is later
                used to identify each individual query in the result dict.

        Raises:
            ValueError: If no query dictionary is provided.
            TypeError: If any of the arguments provided are the wrong type.

        """
        # first define a function that we're going to use to fetch url results.
        def fetch_json_object(url):
            # Naively try to fetch a JSON result from the given url and return
            # a parsed python object. Let the FuncWrap handle any exceptions.
            res = urllib2.urlopen(url)
            return json.loads(res.read())
        
        # validation and initialisation is done inline here.
        if not queries:
            raise ValueError('Must provide at least one pending query! Got none!')
        _validate_type('num_worker_threads', num_worker_threads, types.IntType) 
        _validate_type('queries', queries, types.DictType)
        
        self.func_wraps = []
        for q_name, q_url in queries.iteritems():
            try:
                _validate_type('', q_name, types.StringType)
                _validate_type('', q_url, types.StringType)
            except TypeError:
                raise TypeError('queries must be a dict of "str_query_name/id": "str_url_to_fetch". ' + 
                    'Instead, got %s: %s in %s' %(q_name, q_url, queries))

            self.func_wraps.append(
                thr.FuncWrap(target=fetch_json_object, name=q_name, args=(q_url,))
            )
        
        self.pool = thr.FuncWrapPool(num_worker_threads, self.func_wraps)
        self.result = {}

    @property
    def is_finished(self):
        """True iff all jobs have finished, regardless of whether there was an exception.

        Returns: 
            bool: Indicates whether all requests have returned either an exception or some result.

        """
        return bool(self.result)

    @property
    def has_exception(self):
        """True iff all jobs have finished and any one of them encountered an exception.

        Returns:
            bool: Indicates whether an exception was raised whilst running any of the requests.

        """
        if not self.result:
            return False
        return any(isinstance(r, Exception) for r in self.result.values())

    def run(self):
        """Run all queries in the pool, await results and collect them in self.result dict.

        When each worker returns (or exits following an exception), the return value (or exception)
        is collected in the self.result dict, which is of the form { 'query_str_label': result_obj }

        Note that the results will automatically be parsed from JSON into python objects.

        Returns:
            None: The return value (or exception) of all queries is stored in the self.result dict.

        """
        self.pool.run()
        for fw in self.func_wraps:
            if fw.exception:
                self.result[fw.name] = fw.exception
            else:
                self.result[fw.name] = fw.result
            



__author__ = 'nickmab'

import types
import inspect

'''Utility code for use by other py files within this package.'''

def validate_type(obj_name, obj, desired_type):
    if not isinstance(obj, desired_type):
        raise TypeError('%s must be a %s. Instead got %s: %s.'
            %(obj_name, desired_type.__name__, type(obj).__name__), obj)

def is_coroutine(obj):
    if not isinstance(obj, types.FunctionType):
        return False
    try:
        yields = filter(lambda l: 'yield' in l, inspect.getsourcelines(obj)[0])
    except IOError:
        # indicates could not get source code for the object.
        return None
    for line in yields:
        if '=' not in line:
            # must be used like "x = yield" and a yield on its own is considered invalid.
            return False
        if 'yield' in line.split('=')[0]:
            # catch edge cases like "yield x == y"
            return False
    return True
